// SPDX-License-Identifier: BSD-3-Clause

#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include "ring_buffer.h"
#include "utils.h"

// struct to store the sizes of the two parts of the buffer when splitting
typedef struct {
	size_t first_part;
	size_t second_part;
} SplitSizes;

SplitSizes split_space(size_t size, size_t space_at_end)
{
	SplitSizes sizes;
	// split the space into two parts if the size is bigger than the space at the end
	if (size > space_at_end) {
		sizes.first_part = space_at_end;
		sizes.second_part = size - space_at_end;
	} else {
		// otherwise it's all in the first part
		sizes.first_part = size;
		sizes.second_part = 0;
	}
	return sizes;
}

static size_t get_free_space(const so_ring_buffer_t *ring)
{
	// calculate the free space in the ring buffer by subtracting the length from the capacity
	return ring->cap - ring->len;
}

static size_t adjust_position(size_t pos, size_t offset, size_t cap)
{
	// adjust the position by adding the offset
	size_t new_pos = pos + offset;
	// if the new position is greater than the capacity, wrap around
	if (new_pos >= cap)
		new_pos -= cap;
	return new_pos;
}

static void copy_to_ring(so_ring_buffer_t *ring, const void *data, size_t size)
{
	// calculate the space at the end of the buffer
	size_t space_at_end = ring->cap - ring->write_pos;
	// split the space into two parts
	SplitSizes sizes = split_space(size, space_at_end);

	// copy the data to the buffer
	memcpy(ring->data + ring->write_pos, data, sizes.first_part);
	// if there is a second part, copy it to the beginning of the buffer
	if (sizes.second_part > 0)
		memcpy(ring->data, (const char *)data + sizes.first_part, sizes.second_part);
}

static void copy_from_ring(so_ring_buffer_t *ring, void *data, size_t size)
{
	// calculate the space at the end of the buffer
	size_t space_at_end = ring->cap - ring->read_pos;
	// split the space into two parts
	SplitSizes sizes = split_space(size, space_at_end);

	// copy the data from the buffer , no need to check for second part
	memcpy(data, ring->data + ring->read_pos, sizes.first_part);
}

int ring_buffer_init(so_ring_buffer_t *ring, size_t cap)
{
	ring->data = malloc(cap); // allocate memory for the ring buffer
	if (ring->data == NULL)
		DIE(1, "malloc failed");

	// initialize the ring buffer
	ring->cap = cap;
	ring->len = 0;
	ring->read_pos = 0;
	ring->write_pos = 0;
	ring->stop = 0;

	// initialize the synchronization primitives for synchronization
	int result = pthread_mutex_init(&ring->mutex, NULL);

	if (result != 0)
		DIE(1, "pthread_mutex_init failed");

	result = pthread_cond_init(&ring->not_empty, NULL);
	if (result != 0)
		DIE(1, "failed to initialize not_empty");

	result = pthread_cond_init(&ring->not_full, NULL);
	if (result != 0)
		DIE(1, "failed to initialize not_full");
	return 0;
}

ssize_t ring_buffer_enqueue(so_ring_buffer_t *ring, void *data, size_t size)
{
	if (data == NULL) {
		ERR(1, "invalid data pointer");
		return -1;
	}

	// lock to ensure exclusive access to the ring buffer
	int result = pthread_mutex_lock(&ring->mutex);

	if (result != 0) {
		ERR(result != 0, "pthread_mutex_lock failed");
		return -1;
	}

	// wait until there is enough space in the buffer
	while (get_free_space(ring) < size && !ring->stop) {
		result = pthread_cond_wait(&ring->not_full, &ring->mutex);
		if (result != 0) {
			pthread_mutex_unlock(&ring->mutex);
			ERR(result != 0, "failed to wait on not_full condition");
			return -1;
		}
	}

	// if the buffer is stopped and there is no space, return
	if (ring->stop) {
		pthread_mutex_unlock(&ring->mutex);
		return -1;
	}

	// copy the data to the buffer
	copy_to_ring(ring, data, size);

	// update the write position and length
	ring->write_pos = adjust_position(ring->write_pos, size, ring->cap);
	ring->len += size;

	// signal that the buffer is not empty
	result = pthread_cond_signal(&ring->not_empty);
	if (result != 0)
		ERR(result != 0, "failed to signal not_empty condition");

	// unlock the mutex and allow other threads to access the buffer
	result = pthread_mutex_unlock(&ring->mutex);
	if (result != 0)
		ERR(result != 0, "pthread_mutex_unlock failed");
	return size;
}

ssize_t ring_buffer_dequeue(so_ring_buffer_t *ring, void *data, size_t size)
{
	if (data == NULL) {
		ERR(1, "invalid data");
		return -1;
	}

	// lock mutex to ensure exclusive access to the ring buffer
	int result = pthread_mutex_lock(&ring->mutex);

	if (result != 0) {
		ERR(result != 0, "pthread_mutex_lock failed");
		return -1;
	}

	// wait until there is enough data in the buffer
	while (ring->len < size && !ring->stop) {
		result = pthread_cond_wait(&ring->not_empty, &ring->mutex);
		if (result != 0) {
			pthread_mutex_unlock(&ring->mutex);
			ERR(result != 0, "failed to wait on not_empty condition");
			return -1;
		}
	}

	// if the buffer is stopped and there is no data, return
	if (ring->stop && ring->len == 0) {
		pthread_mutex_unlock(&ring->mutex);
		return 0;
	}

	// copy the data from the buffer
	copy_from_ring(ring, data, size);

	// update the read position and length
	ring->read_pos = adjust_position(ring->read_pos, size, ring->cap);
	ring->len -= size;

	// signal that the buffer is not full
	result = pthread_cond_signal(&ring->not_full);
	if (result != 0)
		ERR(result != 0, "failed to signal not_full condition");

	// unlock the mutex and allow other threads to access the buffer
	result = pthread_mutex_unlock(&ring->mutex);
	if (result != 0)
		ERR(result != 0, "pthread_mutex_unlock failed");

	return size;
}

void ring_buffer_destroy(so_ring_buffer_t *ring)
{
	if (ring == NULL || ring->data == NULL)
		DIE(1, "invalid ring buffer");

	// free the data allocated for the ring buffer
	free(ring->data);

	// destroy the synchronization primitives and condition variables
	int result = pthread_mutex_destroy(&ring->mutex);

	if (result != 0)
		DIE(1, "failed to destroy mutex");

	result = pthread_cond_destroy(&ring->not_empty);
	if (result != 0)
		DIE(1, "failed to destroy not_empty");

	result = pthread_cond_destroy(&ring->not_full);
	if (result != 0)
		DIE(1, "failed to destroy not_full");
}

void ring_buffer_stop(so_ring_buffer_t *ring)
{
	if (ring == NULL)
		DIE(1, "invalid ring buffer");

	// lock the motex to modify the stop flag
	int result = pthread_mutex_lock(&ring->mutex);

	if (result != 0)
		DIE(1, "pthread_mutex_lock failed");

	// set the stop flag to 1 to indicate that the buffer is stopped
	ring->stop = 1;

	// wake up all threads waiting on the condition variables and consumers
	result = pthread_cond_broadcast(&ring->not_empty);
	if (result != 0)
		DIE(1, "failed to broadcast not_empty condition");

	result = pthread_cond_broadcast(&ring->not_full);
	if (result != 0)
		DIE(1, "failed to broadcast not_full condition");

	// unlock the mutex to allow other threads to access the buffer
	result = pthread_mutex_unlock(&ring->mutex);
	if (result != 0)
		DIE(1, "pthread_mutex_unlock failed");
}
